import ij.*;
import ij.gui.*;
import ij.util.*;
import ij.plugin.filter.*;
import ij.plugin.filter.PlugInFilter;
import ij.process.*;
import ij.measure.*;
import ij.measure.Calibration;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import ij.text.*;
import multi_plot.*;
import generic_dialog.*;


/** This plugin is for the calculatiom of an azimuthal average.
 *  It lets you choose the starting angle and integration angle
 *  over which the integration on the defined circle is done.
 *  The integration will be done over an area defined by :
 *  starting angle +/- integration angle.
 *  The size and the position of the Roi can be defined and modified by
 *  either using the plugin menu or shortkeys on keyboard.
 *  Additionnally, the integration calculation can be done over a whole stack.
 *
 *  The plugin includes also a routine abling the calculation of the radius
 *  of the generated Roi and saving the data in a result panel.
 *
 *  Last updated: 09-09-2007
 *  author : Philippe Carl (philippe.carl@tiscali.fr)
*/

public class Azimuthal_Average implements PlugInFilter, ActionListener, ItemListener, KeyListener, Measurements
{
	ImagePlus imp;
	ImageProcessor ip2;
	ImageCanvas canvas;
	Rectangle rct;
//	PlotWindowExt pw;
	MultyPlotExt plot;
	static boolean useCalibration = true;
	int nBins = 100;
	int[] xPoint = new int[3];
	int[] yPoint = new int[3];
	int Sa;		// Starting angle in degree over which the calculation is done
	int Ia;		// Integration angle in degree over which the calculation is done
	int Nb;		// Number of sector over which the calculation is done
	double X;	// X center in pixels of the circle over which the calculation is done
	double Y;	// Y center in pixels of the circle over which the calculation is done
	double Ro;	// Radius in pixels of the circle over which the calculation is done
	double Ri;	// Radius in pixels of the inner circle over which the calculation is done
	double cosMin, cosMax, sinMin, sinMax, cosVal, sinVal, angle;
	Button button1, button2, button3, button4, button5, button6;
	Checkbox cb;
	TextField t1 = new TextField( "" );
	TextField t2 = new TextField( "" );
	TextField t3 = new TextField( "" );
	TextField t4 = new TextField( "" );
	TextField t5 = new TextField( "" );
	TextField t6 = new TextField( "" );
	TextField t7 = new TextField( "" );

	public int setup(String arg, ImagePlus imp)
	{
		if (arg.equals("about"))
			{showAbout(); return DONE;}
		this.imp = imp;
		return DOES_ALL+NO_UNDO+ROI_REQUIRED;
	}

	public void run(ImageProcessor ip)
	{
		imp.unlock();
		ip2 = ip;
		Sa = 0;
		Ia = 180;
		Nb = 10;
		ImageWindow win = imp.getWindow();
		canvas = win.getCanvas();
		canvas.addKeyListener(this);
		setXYcenter();
		IJ.makeOval((int)(X-Ro), (int)(Y-Ro), (int)(2*Ro), (int)(2*Ro));
		doDialog();
	}

	public void setXYcenter()
	{
		rct = imp.getRoi().getBoundingRect();
		X   = (double) rct.x + (double) rct.width / 2;
		Y   = (double) rct.y + (double) rct.height / 2;
		Ro  = (rct.width + rct.height) / 4.0;
		Ri  = Ro / 2.0;
	}

	public void updateXY()
	{
		X  = Double.valueOf(t1.getText()).doubleValue();
		Y  = Double.valueOf(t2.getText()).doubleValue();
		Ro = Double.valueOf(t3.getText()).doubleValue();
		Ri = Double.valueOf(t4.getText()).doubleValue();
		Sa = Double.valueOf(t5.getText()).intValue();
		Ia = Double.valueOf(t6.getText()).intValue();
		Nb = Double.valueOf(t7.getText()).intValue();
	}

	private void setValues()
	{
		t1.setText("" + X);
		t2.setText("" + Y);
		t3.setText("" + Ro);
		t4.setText("" + Ri);
		t5.setText("" + Sa);
		t6.setText("" + Ia);
		t7.setText("" + Nb);
	}

	public double cos(double val)
	{
		if(Math.IEEEremainder(val, 2.0 * Math.PI) == 0.0)
			return 1.0;
		else if(Math.IEEEremainder(val, Math.PI) == 0.0)
			return -1.0;
		else if(Math.IEEEremainder(val, Math.PI / 2.0) == 0.0)
			return 0.0;
		else
			return Math.cos(val);
	}

	public double getAngle(double cos, double sin)
	{
		if(sin >= 0)
			return (Math.acos(cos) * 180.0 / Math.PI);
		else
 			return ((2 * Math.PI - Math.acos(cos)) * 180.0 / Math.PI);
	}

	public int colorX(int val)
	{
		if(Math.IEEEremainder((double) (val-1), 4.0) == 0.0)
			return 1;
		else
			return 0;
	}

	public int colorY(int val)
	{
		if(Math.IEEEremainder((double) (val-2), 4.0) == 0.0)
			return 1;
		else
			return 0;
	}

	public int colorZ(int val)
	{
		if(Math.IEEEremainder((double) (val-3), 4.0) == 0.0)
			return 1;
		else
			return 0;
	}

	public double sin(double val)
	{
		if(Math.IEEEremainder(val, Math.PI) == 0.0)
			return 0.0;
		else
			return Math.sin(val);
	}

	private void setCosSin()
	{
		int i;
		double tmpVal;
		cosMin = cos(Math.PI * (Sa - Ia) / 180.0);
		cosMax = cosMin;
		sinMin = sin(Math.PI * (Sa - Ia) / 180.0);
		sinMax = sinMin;
		for(i = 1; i <= 2*Ia; i++)
		{
			tmpVal = cos(Math.PI * (Sa - Ia + i) / 180);
			if(tmpVal > cosMax)
				cosMax = tmpVal;
			else if(tmpVal < cosMin)
				cosMin = tmpVal;
			tmpVal = sin(Math.PI * (Sa - Ia + i) / 180);
			if(tmpVal > sinMax)
				sinMax = tmpVal;
			else if(tmpVal < sinMin)
				sinMin = tmpVal;
		}
	}

	public void doAzimuthalAverage(ImageProcessor ip)
	{
	        String[] headings = new String[2];
		int i;
		int thisBin;
		nBins = (int) Nb;
		float[][] data = new float[2][nBins];
		double a, b;
		double R;
		double xmin = X - Ro, xmax = X + Ro, ymin = Y - Ro, ymax = Y + Ro;
		setCosSin();

		Sa = (Sa > 180) ? (Sa - 360) : (Sa);
		for (a = xmin; a <= xmax; a++)
		{
			for (b = ymin; b <= ymax; b++)
			{
				R = Math.sqrt((a - X) * (a - X) + (b - Y) * (b - Y));
				cosVal = (a - X) / R;
				sinVal = (Y - b) / R;
				if( cosVal >= cosMin && cosVal <= cosMax )
				{
					if( sinVal >= sinMin && sinVal <= sinMax )
					{
						if(((getAngle(cosVal, sinVal) - (Sa - Ia)) / (2 * Ia)) > 1)
							thisBin = (int) Math.floor(((getAngle(cosVal, sinVal) - 360 - (Sa - Ia)) / (2 * Ia)) * (double) nBins);
						else
							thisBin = (int) Math.floor(((getAngle(cosVal, sinVal) - (Sa - Ia)) / (2 * Ia)) * (double) nBins);
						if (thisBin > nBins - 1) thisBin = nBins - 1;
						if(R >= Ri && R <= Ro && thisBin >= 0)
						{
							data[0][thisBin] = data[0][thisBin] + 1;
							data[1][thisBin] = data[1][thisBin] + ip.getPixelValue((int) a, (int) b);
						}
					}
				}
			}
		}

		for (i = 0; i < nBins; i++)
		{
			data[1][i] = data[1][i] / data[0][i];
			data[0][i] = (float) (Sa + Ia * ((double) (2.0 * i + 1.0) / nBins - 1.0));

		}
//		pw = new PlotWindowExt("Azimuthal Average Plot - "+getImageTitle(), "Radius [pixels]", "Normalized Integrated Intensity [-]",  data[0], data[1]);
		plot = new MultyPlotExt(getImageTitle(), "Angle [degre]", "Normalized Integrated Intensity [-]",  data[0], data[1]);
		headings[0] = "Angle [degre]\t";

//		pw.draw();
		headings[1] = "Normalized Integrated Intensity";
	        MultyPlotWindowExt wnd = plot.show();
        	wnd.setLineHeadings(headings, false);
	}

	public void doStackAzimuthalAverage()
	{
		int i, j;
		int thisBin;
		nBins = (int) Nb;
	        String[] headings = new String[imp.getStackSize()+1];
		float[]   dataX = new float[nBins];
		float[][] dataY = new float[imp.getStackSize()][nBins];
		float minY, maxY;
		double[] extrema;
		double a, b;
		double R;
		double xmin = X - Ro, xmax = X + Ro, ymin = Y - Ro, ymax = Y + Ro;
		setCosSin();

		Sa = (Sa > 180) ? (Sa - 360) : (Sa);
		for (j = 0; j <= imp.getStackSize()-1; j++)
		{
			imp.setSlice(j+1);
			headings[j+1] = getImageTitle();
			for (a = xmin; a <= xmax; a++)
			{
				for (b = ymin; b <= ymax; b++)
				{
					R = Math.sqrt((a - X) * (a - X) + (b - Y) * (b - Y));
					cosVal = (a - X) / R;
					sinVal = (Y - b) / R;
					if( cosVal >= cosMin && cosVal <= cosMax )
					{
						if( sinVal >= sinMin && sinVal <= sinMax )
						{
							if(((getAngle(cosVal, sinVal) - (Sa - Ia)) / (2 * Ia)) > 1)
								thisBin = (int) Math.floor(((getAngle(cosVal, sinVal) - 360 - (Sa - Ia)) / (2 * Ia)) * (double) nBins);
							else
								thisBin = (int) Math.floor(((getAngle(cosVal, sinVal) - (Sa - Ia)) / (2 * Ia)) * (double) nBins);
							if (thisBin > nBins - 1) thisBin = nBins - 1;
							if(R >= Ri && R <= Ro && thisBin >= 0)
							{
								dataX   [thisBin] = dataX   [thisBin] + 1;
								dataY[j][thisBin] = dataY[j][thisBin] + ip2.getPixelValue((int) a, (int) b);
							}
						}
					}
				}
			}
			for (i = 0; i < nBins; i++)
			{
				dataY[j][i] =  dataY[j][i] / dataX[i];
				dataX[i] = 0;
			}
		}
		minY = dataY[0][0];
		maxY = dataY[0][0];
		for (j = 0; j <= imp.getStackSize()-1; j++)
		{
			extrema = Tools.getMinMax(dataY[j]);
			if (extrema[0] < minY)
				minY = (float) extrema[0];
			else if (extrema[1] > maxY)
				maxY = (float) extrema[1];
		}

		for (i = 0; i < nBins; i++)
			dataX[i] = (float) (Sa + Ia * ((double) (2.0 * i + 1.0) / nBins - 1.0));
		plot = new MultyPlotExt("Azimuthal Average Plot", "Angle [degre]", "Normalized Integrated Intensity [-]", dataX, minY, maxY);
		headings[0] = "Angle [degre]";

		for (j = 0; j <= imp.getStackSize()-1; j++)
		{
			plot.setColor(new Color(colorX(j) * 0xff, colorY(j) * 0xff, colorZ(j) * 0xff));
			plot.addPoints(dataX,dataY[j],2);
		}
		plot.setLimits(dataX[0], dataX[nBins-1], minY, maxY);
		MultyPlotWindowExt wnd = plot.show();
		wnd.setLineHeadings(headings, false);
//		wnd.setPrecision(3,3);
	}

	public String getImageTitle()
	{
		String str;
		if (imp.getStackSize()>1)
		{
			ImageStack stack = imp.getStack();
			return stack.getShortSliceLabel(imp.getCurrentSlice());
		}
		else
		{
			str = imp.getTitle();
			int len = str.length();
			if (len>4 && str.charAt(len-4)=='.' && !Character.isDigit(str.charAt(len-1)))
			return str.substring(0,len-4);  
		}
		return "";
	}

	public void calculateRoiRadius(ImageProcessor ip)
	{
		String title;
		Calibration cal = imp.getCalibration();
		int measurements = Analyzer.getMeasurements();			// defined in Set Measurements dialog
		Analyzer.setMeasurements(measurements);
		Analyzer a = new Analyzer();
		ImageStatistics stats = imp.getStatistics(measurements);
		Roi roi = imp.getRoi();
		title = getImageTitle();
		a.saveResults(stats, roi);					// store in system Results table
		ResultsTable rt = Analyzer.getResultsTable();			// get the system Results table
		rt.addLabel("Label", title);
		if (useCalibration)
		{
			rt.addValue("Radius ["+cal.getUnits()+"]", cal.pixelWidth*Ro);
			rt.addValue("Radius [pixels]", Ro);
		}
		else
			rt.addValue("Radius [pixels]", Ro);
//		rt.addValue("X Center [pixels]", X);
//		rt.addValue("Y Center [pixels]", Y);
//		rt.addValue("Starting Angle []", Sa);
//		rt.addValue("Integration Angle []", Ia);
		a.displayResults();						// display the Results in the worksheet
		a.updateHeadings();						// update the worksheet headings
	}

	private void doDialog()
	{
		GenericRecallableDialog gd = new GenericRecallableDialog("Azimuthal Average", IJ.getInstance());
		gd.addPanel(addPanel());
		gd.showDialog();
		while (!(gd.wasCanceled())) { }
		canvas.removeKeyListener(this);
	}

	private Panel smallPanel(String str, TextField tf)
	{
		Panel spanel = new Panel();
//		spanel.setLayout(new GridLayout(1, 2));
		spanel.add( new Label(str) );
		spanel.add( tf );
		return spanel;
	}

	private Panel addPanel()
	{
		Panel panel = new Panel();
		panel.setLayout(new GridLayout(14, 1));
		panel.add(smallPanel("X center (pixels):", t1));
		panel.add(smallPanel("Y center (pixels):", t2));
		panel.add(smallPanel("Outer Radius (pixels):  ", t3));
		button1 = new Button("Plot Outer Radius (g)");
		button1.addActionListener(this);
		panel.add(button1);
		panel.add(smallPanel("Inner Radius (pixels):  ", t4));
		button2 = new Button("Plot Inner Radius (j)");
		button2.addActionListener(this);
		panel.add(button2);
		panel.add(smallPanel("Starting angle ():     ", t5));
		panel.add(smallPanel("Integration angle ():", t6));
		button3 = new Button("Plot Integration Area (l)");
		button3.addActionListener(this);
		panel.add(button3);
		panel.add(smallPanel("Number of bins:", t7));
		button4 = new Button("Calculate ROI Radius (u)");
		button4.addActionListener(this);
		panel.add(button4);
		button5 = new Button("Calculate Azimuthal Average");
		button5.addActionListener(this);
		panel.add(button5);
		button6 = new Button("Calculate Stack Azimuthal Average");
		button6.addActionListener(this);
		panel.add(button6);
		cb = new Checkbox("Use Spatial Calibration"); 
		cb.setState(useCalibration);
		cb.addItemListener (this);
		panel.add(cb);
		setValues();

		return panel;
	}

	public synchronized void itemStateChanged(ItemEvent e) { }

	public void actionPerformed(ActionEvent e)
	{
		Object b = e.getSource();

		if (b==button1)
		{
			updateXY();
			imp.setRoi(new OvalRoi((int)(X-Ro), (int)(Y-Ro), (int)(2*Ro), (int)(2*Ro)));
			imp.repaintWindow();
		}
		if (b==button2)
		{
			updateXY();
			imp.setRoi(new OvalRoi((int)(X-Ri), (int)(Y-Ri), (int)(2*Ri), (int)(2*Ri)));
			imp.repaintWindow();
		}
		if (b==button3)
		{
			updateXY();
			xPoint[0] = (int) ( X + Ro * cos(Math.PI * (Sa - Ia) / 180) );
			xPoint[1] = (int) ( X );
			xPoint[2] = (int) ( X + Ro * cos(Math.PI * (Sa + Ia) / 180) );
			yPoint[0] = (int) ( Y - Ro * sin(Math.PI * (Sa - Ia) / 180) );
			yPoint[1] = (int) ( Y );
			yPoint[2] = (int) ( Y - Ro * sin(Math.PI * (Sa + Ia) / 180) );
			imp.setRoi(new PolygonRoi(xPoint, yPoint, 3,  Roi.ANGLE));
			imp.repaintWindow();
		}
		if (b==button4)
		{
			useCalibration = cb.getState();
			updateXY();
			calculateRoiRadius(ip2);
		}
		if (b==button5)
		{
			updateXY();
			doAzimuthalAverage(ip2);
		}
		if (b==button6)
		{
			if (imp.getStackSize()>1)
			{
				updateXY();
				doStackAzimuthalAverage();
			}
			else
				IJ.showMessage("Error", "Stack required");
		}
	}

	public void keyPressed(KeyEvent e)
	{
		int keyCode = e.getKeyCode();
		int flags   = e.getModifiers();

		if (keyCode == KeyEvent.VK_RIGHT || keyCode == KeyEvent.VK_NUMPAD6)
		{
			if (flags==KeyEvent.ALT_MASK)
				X+=5;
			else if (flags==KeyEvent.CTRL_MASK)
				X+=10;
			else
				X++;
			t1.setText(""+X);
			imp.setRoi(new OvalRoi((int)(X-Ro), (int)(Y-Ro), (int)(2*Ro), (int)(2*Ro)));
			imp.repaintWindow();			
		}
		if (keyCode == KeyEvent.VK_LEFT || keyCode == KeyEvent.VK_NUMPAD4)
		{
			if (flags==KeyEvent.ALT_MASK)
				X-=5;
			else if (flags==KeyEvent.CTRL_MASK)
				X-=10;
			else
				X--;
			t1.setText(""+X);
			imp.setRoi(new OvalRoi((int)(X-Ro), (int)(Y-Ro), (int)(2*Ro), (int)(2*Ro)));
			imp.repaintWindow();			
		}
		if (keyCode == KeyEvent.VK_DOWN || keyCode == KeyEvent.VK_NUMPAD2)
		{
			if (flags==KeyEvent.ALT_MASK)
				Y+=5;
			else if (flags==KeyEvent.CTRL_MASK)
				Y+=10;
			else
				Y++;
			t2.setText(""+Y);
			imp.setRoi(new OvalRoi((int)(X-Ro), (int)(Y-Ro), (int)(2*Ro), (int)(2*Ro)));
			imp.repaintWindow();			
		}
		if (keyCode == KeyEvent.VK_UP || keyCode == KeyEvent.VK_NUMPAD8)
		{
			if (flags==KeyEvent.ALT_MASK)
				Y-=5;
			else if (flags==KeyEvent.CTRL_MASK)
				Y-=10;
			else
				Y--;
			t2.setText(""+Y);
			imp.setRoi(new OvalRoi((int)(X-Ro), (int)(Y-Ro), (int)(2*Ro), (int)(2*Ro)));
			imp.repaintWindow();			
		}
		if (keyCode == KeyEvent.VK_PAGE_UP || keyCode == e.VK_ADD)
		{
			if (flags==KeyEvent.ALT_MASK)
				Ro+=5;
			else if (flags==KeyEvent.CTRL_MASK)
				Ro+=10;
			else
				Ro++;
			t3.setText(""+Ro);
			imp.setRoi(new OvalRoi((int)(X-Ro), (int)(Y-Ro), (int)(2*Ro), (int)(2*Ro)));
			imp.repaintWindow();
		}
		if (keyCode == KeyEvent.VK_PAGE_DOWN || keyCode == e.VK_SUBTRACT)
		{
			if (flags==KeyEvent.ALT_MASK)
			{
				if(Ro >= 5)
					Ro-=5;
			}
			else if (flags==KeyEvent.CTRL_MASK)
			{
				if(Ro >= 5)
					Ro-=10;
			}
			else
			{
				if(Ro >= 1)
					Ro--;
			}
			t3.setText(""+Ro);
			if(Ri > Ro)
			{
				Ri = Ro;
				t4.setText(""+Ri);
			}
			imp.setRoi(new OvalRoi((int)(X-Ro), (int)(Y-Ro), (int)(2*Ro), (int)(2*Ro)));
			imp.repaintWindow();
		}
		if (keyCode == KeyEvent.VK_HOME)
		{
			if (flags==KeyEvent.ALT_MASK)
				Ri+=5;
			else if (flags==KeyEvent.CTRL_MASK)
				Ri+=10;
			else
				Ri++;
			t4.setText(""+Ri);
			if(Ri > Ro)
			{
				Ro = Ri;
				t3.setText(""+Ro);
			}
			imp.setRoi(new OvalRoi((int)(X-Ri), (int)(Y-Ri), (int)(2*Ri), (int)(2*Ri)));
			imp.repaintWindow();
		}
		if (keyCode == KeyEvent.VK_END)
		{
			if (flags==KeyEvent.ALT_MASK)
			{
				if(Ri >= 5)
					Ri-=5;
			}
			else if (flags==KeyEvent.CTRL_MASK)
			{
				if(Ri >= 5)
					Ri-=10;
			}
			else
			{
				if(Ri >= 1)
					Ri--;
			}
			t4.setText(""+Ri);
			imp.setRoi(new OvalRoi((int)(X-Ri), (int)(Y-Ri), (int)(2*Ri), (int)(2*Ri)));
			imp.repaintWindow();
		}
		if (keyCode == e.VK_G)
		{
			updateXY();
			imp.setRoi(new OvalRoi((int)(X-Ro), (int)(Y-Ro), (int)(2*Ro), (int)(2*Ro)));
			imp.repaintWindow();
		}
		if (keyCode == e.VK_J)
		{
			updateXY();
			imp.setRoi(new OvalRoi((int)(X-Ri), (int)(Y-Ri), (int)(2*Ri), (int)(2*Ri)));
			imp.repaintWindow();
		}
		if (keyCode == e.VK_L)
		{
			updateXY();
			xPoint[0] = (int) ( X + Ro * cos(Math.PI * (Sa - Ia) / 180) );
			xPoint[1] = (int) ( X );
			xPoint[2] = (int) ( X + Ro * cos(Math.PI * (Sa + Ia) / 180) );
			yPoint[0] = (int) ( Y - Ro * sin(Math.PI * (Sa - Ia) / 180) );
			yPoint[1] = (int) ( Y );
			yPoint[2] = (int) ( Y - Ro * sin(Math.PI * (Sa + Ia) / 180) );
			imp.setRoi(new PolygonRoi(xPoint, yPoint, 3,  Roi.ANGLE));
			imp.repaintWindow();
		}
		if (keyCode == e.VK_U)
		{
			useCalibration = cb.getState();
			updateXY();
			calculateRoiRadius(ip2);
		}
//		if (keyCode == e.VK_Q)
//		{
//			updateXY();
//			doAzimuthalAverage(ip2);
//		}
//		if (keyCode == e.VK_U)
//		{
//			if (imp.getStackSize()>1)
//			{
//				updateXY();
//				doStackAzimuthalAverage();
//			}
//			else
//				IJ.showMessage("Error", "Stack required");
//		}
	}

	public void keyReleased(KeyEvent e) { }

	public void keyTyped(KeyEvent e) { }

	public void showAbout()
	{
		IJ.showMessage("Azimuthal Average",
			"This plugin is for the calculatiom of an azimuthal average.\n" +
			"The plugin lets you choose the starting angle and integration angle over\n" +
			"which the integration on the defined Roi is done.\n" +
			"The integration will be done over an area defined by :\n" +
			"starting angle +/- integration angle.\n" +
			"The size and the position of the Roi can be defined and modified by\n" +
			"either using the plugin menu or shortkeys on keyboard.\n" +
			"Additionnally, the integration calculation can be done over a whole stack.\n" +
			"                                                                                                                                               \n" +
			"The plugin includes also a routine abling the calculation of the radius\n" +
			"of the generated Roi and saving the data in a Result panel.\n" +
			"                                                                                                                                               \n" +
			"Last updated: 09-09-2007\n\n" +
			"Author : Philippe Carl (philippe.carl@tiscali.fr)"
		);
	}

}
